@@ -1,13 +1,10 @@
 # This file is part of Parti.
 # Copyright (C) 2011 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
-# Copyright (C) 2010-2012 Antoine Martin <antoine@devloop.org.uk>
+# Copyright (C) 2010-2013 Antoine Martin <antoine@devloop.org.uk>
 # Copyright (C) 2008, 2010 Nathaniel Smith <njs@pobox.com>
 # Parti is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
-#disabled for now as this causes bugs:
-PRESERVE_WORSPACE = False
-
 #pygtk3 vs pygtk2 (sigh)
 from wimpiggy.gobject_compat import import_gobject, import_gtk, import_gdk, is_gtk3
 gobject = import_gobject()
@@ -77,8 +74,14 @@ else:
         gtk.Window.__init__(gtkwindow, wintype)
     def is_mapped(win):
         return win.window is not None and win.window.is_visible()
+    if gtk.gtk_version>=(2,14):
+        def gdk_window(gtkwindow):
+            return gtkwindow.get_window()
+    else:
+        def gdk_window(gtkwindow):
+            return gtkwindow.window
     def get_window_geometry(gtkwindow):
-        gdkwindow = gtkwindow.get_window()
+        gdkwindow = gdk_window(gtkwindow)
         x, y = gdkwindow.get_origin()
         _, _, w, h, _ = gdkwindow.get_geometry()
         return (x, y, w, h)
@@ -86,7 +89,7 @@ else:
         gtkwindow.set_geometry_hints(None, **hints)
 
     def queue_draw(gtkwindow, x, y, width, height):
-        window = gtkwindow.get_window()
+        window = gdk_window(gtkwindow)
         if window:
             window.invalidate_rect(gdk.Rectangle(x, y, width, height), False)
         else:
@@ -128,6 +131,7 @@ else:
                   }
 
 
+import os
 import cairo
 import re
 import sys
@@ -135,13 +139,45 @@ import sys
 from wimpiggy.log import Logger
 log = Logger()
 
-from xpra.window_backing import new_backing
+DRAW_DEBUG = os.environ.get("XPRA_DRAW_DEBUG", "0")=="1"
+
 try:
-    from wimpiggy.prop import prop_set, prop_get
+    from wimpiggy.prop import prop_get
     has_wimpiggy_prop = True
 except ImportError, e:
     has_wimpiggy_prop = False
 
+def xget_u32_property(target, name):
+    try:
+        if not has_wimpiggy_prop:
+            if is_gtk3():
+                name_atom = gdk.Atom.intern(name, False)
+                type_atom = gdk.Atom.intern("CARDINAL", False)
+                gdk.property_get(target, name_atom, type_atom, 0, 9999, False)
+            else:
+                prop = target.property_get(name)
+            if not prop or len(prop)!=3 or len(prop[2])!=1:
+                return  None
+            log("xget_u32_property(%s, %s)=%s", target, name, prop[2][0])
+            return prop[2][0]
+        v = prop_get(target, name, "u32", ignore_errors=True)
+        log("xget_u32_property(%s, %s)=%s", target, name, v)
+        if type(v)==int:
+            return  v
+    except Exception, e:
+        log.error("xget_u32_property error on %s / %s: %s", target, name, e)
+    return None
+
+CAN_SET_WORKSPACE = False
+if not sys.platform.startswith("win") and has_wimpiggy_prop:
+    try:
+        #TODO: in theory this is not a proper check, meh - that will do
+        root = gtk.gdk.get_default_root_window()
+        supported = prop_get(root, "_NET_SUPPORTED", ["atom"], ignore_errors=True)
+        CAN_SET_WORKSPACE = bool(supported) and "_NET_WM_DESKTOP" in supported
+    except:
+        pass
+
 if sys.version < '3':
     import codecs
     def u(x):
@@ -153,12 +189,13 @@ else:
 
 
 class ClientWindow(gtk.Window):
-    def __init__(self, client, wid, x, y, w, h, metadata, override_redirect, client_properties, auto_refresh_delay):
+    def __init__(self, client, group_leader, wid, x, y, w, h, metadata, override_redirect, client_properties, auto_refresh_delay):
         if override_redirect:
             init_window(self, WINDOW_POPUP)
         else:
             init_window(self, WINDOW_TOPLEVEL)
         self._client = client
+        self.group_leader = group_leader
         self._id = wid
         self._pos = (-1, -1)
         self._size = (1, 1)
@@ -190,107 +227,53 @@ class ClientWindow(gtk.Window):
             if transient_for is not None and transient_for.window is not None and type_hint in OR_TYPE_HINTS:
                 transient_for._override_redirect_windows.append(self)
         self.connect("notify::has-toplevel-focus", self._focus_change)
+        #deal with workspace mapping:
 
-    def do_realize(self):
-        if not PRESERVE_WORSPACE:
-            gtk.Window.do_realize(self)
+    def set_workspace(self):
+        if not CAN_SET_WORKSPACE or self._been_mapped:
+            return
+        workspace = self._client_properties.get("workspace")
+        log("set_workspace() workspace=%s", workspace)
+        if not workspace or workspace==self.get_workspace():
             return
-        ndesktops = 0
         try:
-            root = gtk.gdk.screen_get_default().get_root_window()
-            prop = root.property_get("_NET_NUMBER_OF_DESKTOPS")
-            if prop is not None and len(prop)==3 and len(prop[2])==1:
-                ndesktops = prop[2][0]
-        except Exception, e:
-            log.error("failed to get workspace count: %s", e)
-        workspace = self._client_properties.get("workspace", -1)
-        log("do_realize() workspace=%s (ndesktops=%s)", workspace, ndesktops)
-
-        #below we duplicate gtk.window.realize() code
-        #just so we can insert the property code at the right place:
-        #after the gdk.Window is created, but before it gets positionned.
-        allocation = self.get_allocation()
-        if allocation.x==-1 and allocation.y==-1 and allocation.width==1 and allocation.height==1:
-            w, h = self.size_request()
-            if w>0 or h>0:
-                allocation.width = w
-                allocation.height = h
-            self.size_allocate(w, h)
-            self.queue_resize()
-            if self.flags() & gtk.REALIZED:
-                log.error("window is already realized!")
+            from wimpiggy.lowlevel import sendClientMessage, const  #@UnresolvedImport
+            from wimpiggy.error import trap
+            root = gdk_window(self).get_screen().get_root_window()
+            ndesktops = xget_u32_property(root, "_NET_NUMBER_OF_DESKTOPS")
+            log("set_workspace() ndesktops=%s", ndesktops)
+            if ndesktops is None or ndesktops<=1:
                 return
+            workspace = max(0, min(ndesktops-1, workspace))
+            event_mask = const["SubstructureNotifyMask"] | const["SubstructureRedirectMask"]
+            trap.call_synced(sendClientMessage, root, gdk_window(self), False, event_mask, "_NET_WM_DESKTOP",
+                      workspace, const["CurrentTime"],
+                      0, 0, 0)
+        except Exception, e:
+            log.error("failed to set workspace: %s", e)
 
-        self.set_flags(gtk.REALIZED)
-        is_toplevel = self.get_parent() is None
-        if hasattr(self, "is_toplevel"):
-            is_toplevel = self.is_toplevel()
-        if is_toplevel:
-            window_type = gtk.gdk.WINDOW_TOPLEVEL
-        else:
-            window_type = gtk.gdk.WINDOW_TEMP
-        if self.get_has_frame():
-            #TODO: duplicate gtk code here too..
-            pass
+    def is_OR(self):
+        return self._override_redirect
 
-        events = self.get_events() | gdk.EXPOSURE_MASK | gdk.STRUCTURE_MASK | \
-                    gdk.KEY_PRESS_MASK | gdk.KEY_RELEASE_MASK
-        self.window = gdk.Window(
-            self.get_root_window(),
-            x=allocation.x, y=allocation.y, width=allocation.width, height=allocation.height,
-            wmclass_name=self.wmclass_name, wmclass_class=self.wmclass_class,
-            window_type=window_type,
-            wclass=gdk.INPUT_OUTPUT,
-            title=self.get_title(),
-            event_mask=events,
-            )
+    def is_tray(self):
+        return False
 
-        if has_wimpiggy_prop and not self._override_redirect and ndesktops>workspace and workspace>=0:
-            try:
-                prop_set(self.window, "_NET_WM_DESKTOP", "u32", workspace)
-            except Exception, e:
-                log.error("failed to set workspace: %s", e)
-
-        self.window.set_opacity(1.0)
-        #self.window.enable_synchronized_configure().. not used?
-        self.window.set_user_data(self)
-        self.style.attach(self.window)
-        self.style.set_background(self.window, gtk.STATE_NORMAL)
-        #self.paint() does not exist in pygtk..
-        transient_for = self.get_transient_for()
-        if transient_for and transient_for.flags() & gtk.REALIZED:
-            self.window.set_transient_for(transient_for.get_window())
-        if not self.get_decorated():
-            self.window.set_decorations(0)
-        if not self.get_deletable():
-            self.window.set_functions(gtk.gdk.FUNC_ALL | gtk.gdk.FUNC_CLOSE)
-        if self.get_skip_pager_hint():
-            self.window.set_skip_pager_hint(True)
-        if self.get_skip_taskbar_hint():
-            self.window.set_skip_taskbar_hint(True)
-        self.window.set_accept_focus(self.get_accept_focus())
-        self.window.set_focus_on_map(self.get_focus_on_map())
-        self.window.set_modal_hint(self.get_modal())
-        #cannot access startup id...
+    def is_GL(self):
+        return False
 
     def get_workspace(self):
-        try:
-            if sys.platform.startswith("win"):
-                return  -1              #windows does not have workspaces
-            if not has_wimpiggy_prop:
-                prop = self.window.get_screen().get_root_window().property_get("_NET_CURRENT_DESKTOP")
-                if not prop or len(prop)!=3 or len(prop[2])!=1:
-                    return  -1
-                return prop[2][0]
-            v = prop_get(self.get_window(), "_NET_WM_DESKTOP", "u32", ignore_errors=True)
-            if type(v)==int:
-                return  v
-        except Exception, e:
-            log.error("failed to detect workspace: %s", e)
+        if sys.platform.startswith("win"):
+            return  -1              #windows does not have workspaces
+        window = gdk_window(self)
+        root = window.get_screen().get_root_window()
+        for target, prop in ((window, "_NET_WM_DESKTOP"), (root, "_NET_CURRENT_DESKTOP")):
+            value = xget_u32_property(target, prop)
+            if value is not None:
+                return value
         return  -1
 
-
     def new_backing(self, w, h):
+        from xpra.window_backing import new_backing
         self._backing = new_backing(self._id, w, h, self._backing, self._client.supports_mmap, self._client.mmap)
 
     def update_metadata(self, metadata):
@@ -350,29 +333,19 @@ class ClientWindow(gtk.Window):
             self.set_wmclass(*self._metadata.get("class-instance",
                                                  ("xpra", "Xpra")))
 
+        modal = self._metadata.get("modal", False)
+        self.set_modal(modal or False)
+
         if "icon" in self._metadata:
-            (width, height, coding, data) = self._metadata["icon"]
-            if coding == "premult_argb32":
-                cairo_surf = cairo.ImageSurface(cairo.FORMAT_ARGB32, width, height)
-                cairo_surf.get_data()[:] = data
-                # FIXME: We round-trip through PNG. This is ridiculous, but faster
-                # than doing a bunch of alpha un-premultiplying and byte-swapping
-                # by hand in Python (better still would be to write some Pyrex,
-                # but I don't have time right now):
-                loader = gdk.PixbufLoader()
-                cairo_surf.write_to_png(loader)
-                loader.close()
-                pixbuf = loader.get_pixbuf()
-            else:
-                loader = gdk.PixbufLoader(coding)
-                loader.write(data, len(data))
-                loader.close()
-                pixbuf = loader.get_pixbuf()
-            self.set_icon(pixbuf)
+            width, height, coding, data = self._metadata["icon"]
+            self.update_icon(width, height, coding, data)
 
         if "transient-for" in self._metadata:
             wid = self._metadata.get("transient-for")
-            window = self._client._id_to_window.get(wid)
+            if wid==-1:
+                window = gtk.gdk.get_default_root_window()
+            else:
+                window = self._client._id_to_window.get(wid)
             log("found transient-for: %s / %s", wid, window)
             if window:
                 self.set_transient_for(window)
@@ -388,6 +361,25 @@ class ClientWindow(gtk.Window):
                     self.set_type_hint(hint)
                     break
 
+    def update_icon(self, width, height, coding, data):
+        if coding == "premult_argb32":
+            cairo_surf = cairo.ImageSurface(cairo.FORMAT_ARGB32, width, height)
+            cairo_surf.get_data()[:] = data
+            # FIXME: We round-trip through PNG. This is ridiculous, but faster
+            # than doing a bunch of alpha un-premultiplying and byte-swapping
+            # by hand in Python (better still would be to write some Pyrex,
+            # but I don't have time right now):
+            loader = gdk.PixbufLoader()
+            cairo_surf.write_to_png(loader)
+            loader.close()
+            pixbuf = loader.get_pixbuf()
+        else:
+            loader = gdk.PixbufLoader(coding)
+            loader.write(data, len(data))
+            loader.close()
+            pixbuf = loader.get_pixbuf()
+        self.set_icon(pixbuf)
+
     def refresh_window(self, *args):
         log("refresh_window(%s) wid=%s", args, self._id)
         self._client.send_refresh(self._id)
@@ -399,8 +391,11 @@ class ClientWindow(gtk.Window):
         self._client.send_refresh_all()
 
     def draw_region(self, x, y, width, height, coding, img_data, rowstride, packet_sequence, options, callbacks):
+        """ Note: this runs from the draw thread (not UI thread) """
+        assert self._backing, "window %s has no backing!" % self._id
         def after_draw_refresh(success):
-            log("after_draw_refresh(%s) options=%s", success, options)
+            if DRAW_DEBUG:
+                log.info("after_draw_refresh(%s) options=%s", success, options)
             if success:
                 queue_draw(self, x, y, width, height)
             #clear the auto refresh if enough pixels were sent (arbitrary limit..)
@@ -424,14 +419,16 @@ class ClientWindow(gtk.Window):
 
     """ gtk3 """
     def do_draw(self, context):
-        log("do_draw(%s)", context)
-        if self.get_mapped():
+        if DRAW_DEBUG:
+            log.info("do_draw(%s)", context)
+        if self.get_mapped() and self._backing:
             self._backing.cairo_draw(context, 0, 0)
 
     """ gtk2 """
     def do_expose_event(self, event):
-        log("do_expose_event(%s) area=%s", event, event.area)
-        if not (self.flags() & gtk.MAPPED):
+        if DRAW_DEBUG:
+            log.info("do_expose_event(%s) area=%s", event, event.area)
+        if not (self.flags() & gtk.MAPPED) or self._backing is None:
             return
         x,y,_,_ = event.area
         context = self.window.cairo_create()
@@ -442,10 +439,14 @@ class ClientWindow(gtk.Window):
     def do_map_event(self, event):
         log("Got map event: %s", event)
         gtk.Window.do_map_event(self, event)
+        #set group leader (but avoid ugly "not implemented" warning on win32):
+        if self.group_leader and not sys.platform.startswith("win"):
+            self.window.set_group(self.group_leader)
+        self.set_workspace()
         if not self._override_redirect:
             x, y, w, h = get_window_geometry(self)
             client_properties = {"workspace" : self.get_workspace()}
-            self._client.send(["map-window", self._id, x, y, w, h, client_properties])
+            self._client.send("map-window", self._id, x, y, w, h, client_properties)
             self._pos = (x, y)
             self._size = (w, h)
         self._been_mapped = True
@@ -465,12 +466,12 @@ class ClientWindow(gtk.Window):
         if self._client.window_configure:
             #if we support configure-window, send that first
             client_properties = {"workspace" : self.get_workspace()}
-            self._client.send(["configure-window", self._id, x, y, w, h, client_properties])
+            self._client.send("configure-window", self._id, x, y, w, h, client_properties)
         if dx!=0 or dy!=0:
             #window has moved
             if not self._client.window_configure:
                 #if we don't handle the move via configure:
-                self._client.send(["move-window", self._id, x, y])
+                self._client.send("move-window", self._id, x, y)
             #move any OR window with their parent:
             for window in self._override_redirect_windows:
                 x, y = window.get_position()
@@ -479,7 +480,7 @@ class ClientWindow(gtk.Window):
             self._size = (w, h)
             self.new_backing(w, h)
             if not self._client.window_configure:
-                self._client.send(["resize-window", self._id, w, h])
+                self._client.send("resize-window", self._id, w, h)
 
     def move_resize(self, x, y, w, h):
         assert self._override_redirect
@@ -489,9 +490,13 @@ class ClientWindow(gtk.Window):
         self.new_backing(w, h)
 
     def destroy(self):
+        if self._refresh_timer:
+            gobject.source_remove(self._refresh_timer)
         self._unfocus()
+        if self._backing:
+            self._backing.close()
+            self._backing = None
         gtk.Window.destroy(self)
-        self._backing.close()
 
     def _unfocus(self):
         if self._client._focused==self._id:
@@ -500,10 +505,10 @@ class ClientWindow(gtk.Window):
     def do_unmap_event(self, event):
         self._unfocus()
         if not self._override_redirect:
-            self._client.send(["unmap-window", self._id])
+            self._client.send("unmap-window", self._id)
 
     def do_delete_event(self, event):
-        self._client.send(["close-window", self._id])
+        self._client.send("close-window", self._id)
         return True
 
     def quit(self):
@@ -521,31 +526,35 @@ class ClientWindow(gtk.Window):
     def _pointer_modifiers(self, event):
         pointer = (int(event.x_root), int(event.y_root))
         modifiers = self._client.mask_to_names(event.state)
-        return pointer, modifiers
+        buttons = []
+        for mask, button in {gtk.gdk.BUTTON1_MASK : 1,
+                             gtk.gdk.BUTTON2_MASK : 2,
+                             gtk.gdk.BUTTON3_MASK : 3,
+                             gtk.gdk.BUTTON4_MASK : 4,
+                             gtk.gdk.BUTTON5_MASK : 5}.items():
+            if event.state & mask:
+                buttons.append(button)
+        return pointer, modifiers, buttons
 
     def do_motion_notify_event(self, event):
         if self._client.readonly:
             return
-        (pointer, modifiers) = self._pointer_modifiers(event)
+        pointer, modifiers, buttons = self._pointer_modifiers(event)
         self._client.send_mouse_position(["pointer-position", self._id,
-                                          pointer, modifiers])
+                                          pointer, modifiers, buttons])
 
     def _button_action(self, button, event, depressed):
         if self._client.readonly:
             return
-        (pointer, modifiers) = self._pointer_modifiers(event)
+        pointer, modifiers, buttons = self._pointer_modifiers(event)
         self._client.send_positional(["button-action", self._id,
                                       button, depressed,
-                                      pointer, modifiers])
+                                      pointer, modifiers, buttons])
 
     def do_button_press_event(self, event):
-        if self._client.readonly:
-            return
         self._button_action(event.button, event, True)
 
     def do_button_release_event(self, event):
-        if self._client.readonly:
-            return
         self._button_action(event.button, event, False)
 
     def do_scroll_event(self, event):
