@@ -1,3 +1,10 @@
+#!/usr/bin/env python
+# This file is part of Parti.
+# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
+# Parti is released under the terms of the GNU GPL v2, or, at your option, any
+# later version. See the file COPYING for details.
+
+
 # A utility system for cooperation among stuff that needs nested main-loops.
 #
 # So here's the problem: in the clipboard code, and eventually maybe in other
@@ -53,8 +60,9 @@
 # timeout. If the user ever stops pasting madly for a few seconds, though,
 # then everything should have a chance to return to equilibrium...
 
-import gobject
-import gtk
+from wimpiggy.gobject_compat import import_gobject, import_gtk
+gobject = import_gobject()
+gtk = import_gtk()
 
 from wimpiggy.log import Logger
 log = Logger()
