@@ -1,6 +1,6 @@
 # This file is part of Parti.
 # Copyright (C) 2010 Nathaniel Smith <njs@pobox.com>
-# Copyright (C) 2011, 2012 Antoine Martin <antoine@nagafix.co.uk>
+# Copyright (C) 2011-2013 Antoine Martin <antoine@devloop.org.uk>
 # Parti is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
@@ -8,9 +8,10 @@
 
 import os.path
 
-from xpra.platform.client_extras_base import ClientExtrasBase, WIN32_LAYOUTS
+from xpra.platform.client_extras_base import ClientExtrasBase
 from xpra.platform.clipboard_base import DefaultClipboardProtocolHelper
 from xpra.keys import get_gtk_keymap
+from xpra.platform.keyboard_layouts import WIN32_LAYOUTS
 from wimpiggy.log import Logger
 log = Logger()
 
@@ -27,15 +28,43 @@ class ClientExtras(ClientExtrasBase):
         except ImportError, e:
             log.error("GDK Translated Clipboard failed to load: %s - using default fallback", e)
             self.setup_clipboard_helper(DefaultClipboardProtocolHelper)
-        self.setup_menu()
+        self.setup_exit_handler(True)
+        self.setup_menu(True)
         self.setup_tray(opts.no_tray, opts.notifications, opts.tray_icon)
         self.emulate_altgr = False
         self.last_key_event_sent = None
 
     def cleanup(self):
+        log("cleanup() tray=%s", self.tray)
         ClientExtrasBase.cleanup(self)
         if self.tray:
             self.tray.close()
+            self.tray = None
+
+    def setup_exit_handler(self, enable=1):
+        try:
+            import win32api     #@UnresolvedImport
+            result = win32api.SetConsoleCtrlHandler(self.handle_console_event, enable)
+            if result == 0:
+                log.error("could not SetConsoleCtrlHandler (error %r)", win32api.GetLastError())
+        except:
+            pass
+
+    def handle_console_event(self, event):
+        log("handle_console_event(%s)", event)
+        import win32con         #@UnresolvedImport
+        events = {win32con.CTRL_C_EVENT         : "CTRL_C",
+                  win32con.CTRL_LOGOFF_EVENT    : "SHUTDOWN",
+                  win32con.CTRL_BREAK_EVENT     : "BREAK",
+                  win32con.CTRL_SHUTDOWN_EVENT  : "SHUTDOWN",
+                  win32con.CTRL_CLOSE_EVENT     : "CLOSE"
+                  }
+        if event in events:
+            log.info("exiting on event %s", events.get(event))
+            self.setup_exit_handler(enable=0)       #disable our handler
+            self.quit()
+            return 1
+        return 0
 
     def can_notify(self):
         return  True
@@ -139,11 +168,6 @@ class ClientExtras(ClientExtrasBase):
     def get_gtk_keymap(self):
         return  get_gtk_keymap()
 
-    def grok_modifier_map(self, display_source, xkbmap_mod_meanings):
-        modifiers = ClientExtrasBase.grok_modifier_map(self, display_source, xkbmap_mod_meanings)
-        #modifiers["meta"] = 1 << 3
-        return  modifiers
-
     def get_keymap_modifiers(self):
         """
             ask the server to manage numlock, and lock can be missing from mouse events
