@@ -1,6 +1,6 @@
 /* This file is part of Parti.
  * Copyright (C) 2012 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
- * Copyright (C) 2012 Antoine Martin <antoine@devloop.org.uk>
+ * Copyright (C) 2012, 2013 Antoine Martin <antoine@devloop.org.uk>
  * Parti is released under the terms of the GNU GPL v2, or, at your option, any
  * later version. See the file COPYING for details.
  */
@@ -26,11 +26,22 @@ struct x264lib_ctx;
 /** Expose the encoder pixel format so the decoder can use the same setting */
 int get_encoder_pixel_format(struct x264lib_ctx *ctx);
 
-/** Expose current quality setting so we can tell the client how good the frame was */
+/** Expose current quality setting */
 int get_encoder_quality(struct x264lib_ctx *ctx);
 
+/** Expose current speed setting */
+int get_encoder_speed(struct x264lib_ctx *ctx);
+
+/** Returns the pixel format using our own generic codec_constants */
+int get_pixel_format(int csc);
+
 /** Create an encoding context for images of a given size.  */
-struct x264lib_ctx *init_encoder(int width, int height, int initial_quality, int supports_csc_option);
+struct x264lib_ctx *init_encoder(int width, int height,
+		int initial_quality, int initial_speed,
+		int supports_csc_option,
+		int I422_quality, int I444_quality,
+		int I422_min, int I444_min,
+        char *i420_profile, char *i422_profile, char *i444_profile);
 
 /** Create a decoding context for images of a given size. */
 struct x264lib_ctx *init_decoder(int width, int height, int csc_fmt);
@@ -75,16 +86,15 @@ int csc_image_yuv2rgb(struct x264lib_ctx *ctx, uint8_t *in[3], const int stride[
  @param outsz: Output size
  @param quality_override: Desired quality setting (0 to 100), -1 to use current settings.
 */
-int compress_image(struct x264lib_ctx *ctx, x264_picture_t *pic_in, uint8_t **out, int *outsz, int quality_override);
+int compress_image(struct x264lib_ctx *ctx, x264_picture_t *pic_in, uint8_t **out, int *outsz);
 
 /** Decompress an image using the given context.
  @param in: Input buffer, format is H264.
  @param size: Input size.
- @param out: Will be filled to point to the output data in planar YUV420 format (3 planes). This data will be freed automatically upon next call to the decoder. 
- @param outsize: Output size.
+ @param out: Will be filled to point to the output data in planar YUV420 format (3 planes). This data will be freed automatically upon next call to the decoder.
  @param outstride: Output strides (3 planes).
 */
-int decompress_image(struct x264lib_ctx *, uint8_t *in, int size, uint8_t *(*out)[3], int *outsize, int (*outstride)[3]);
+int decompress_image(struct x264lib_ctx *, uint8_t *in, int size, uint8_t *(*out)[3], int (*outstride)[3]);
 
 /**
  * Change the speed of encoding (x264 preset).
